/*
 * @(#) $Id:  $
 *
 * Copyright 2009/2010 by sIT Solutions,
 * A-1110 Wien, Geiselbergstr.21-25.
 * All rights reserved.
 *
 */
package org.openxma.dsl.platform.i18n;

import java.text.MessageFormat;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.openxma.dsl.platform.exceptions.BeanValidationException;
import org.openxma.dsl.platform.exceptions.ErrorCodedException;


/**
 * Message Provider implementation that resolves message keys (exception error codes) from ResourceBundles.
 * A default resource bundles is always there to be searched.
 * Resource bundles to search can be passed to the constructor or set using the setBundleNames(String[])
 * method. A preferred bundle can be set separately, which will be prepended to the list of the other
 * resource bundles. The order of bundles will always be kept when searching for a message key, using the
 * default bundle last.
 * Take care to only define names of existing resource bundles, otherwise a MissingResourceException might
 * be thrown during message resolution.
 */
public class SimpleErrorMessageProvider implements MessageProvider {

    /** Logger available to subclasses */
    protected final Log logger = LogFactory.getLog(getClass());

    public static String ERROR_MESSAGE_BUNDLE = "ErrorMessage";

    private List<String> bundleNames = new LinkedList<String>();
    private final String defaultBundleName = ErrorCodedException.class.getPackage().getName() + "." + ERROR_MESSAGE_BUNDLE;
    
    public SimpleErrorMessageProvider() {
        super();
    }

    /**
     * Create ErrorMessageProvider, pass bundleNames in order of preference.
     * 
     * @param bundleNames the array of resource bundle names to use for message resolution
     * @see #setBundleNames(String[])
     */
    public SimpleErrorMessageProvider(String[] bundleNames) {
        super();
        setBundleNames(bundleNames);
    }

    /**
     * Return the resolved message for the given key, message parameters and locale from
     * the resource bundles defined for this message provider.
     * The preferred resource bundle will be searched first, then the registered bundles,
     * finally also the default bundle.
     * In case the key is not found the given defaultMessage will be returned.
     * @return the resolved message
     * @throws MissingResourceException in case a registered resource bundle is not found
     */
    public String getMessage(String key, Object[] params, Locale locale, String defaultMessage) {
        String message = null;
        // prepare parameters: at least array of objects should come nicely
        // user defined formatters
        if (params != null) {
            params = prepareParamsForOutput(params);
        }
        for (int i = 0; message == null && i < this.bundleNames.size(); i++) {
            message = resolveMessageForBundle(this.bundleNames.get(i), key, params, locale);
        }
        // TODO: add defaultBundleName to the end of this.bundleNames ?
        if (message == null) {
            message = resolveMessageForBundle(getDefaultBundleName(), key, params, locale);
        }
        return message != null ?
               message :
               defaultMessage;
    }
    
    /**
     * Return the resolved message for the given MessageResolvable,
     * the given resolvable contains all the objects needed to resolve the message.
     * @see #getMessage(String, Object[], Locale, String)
     */
    public String getMessage(MessageResolvable resolvable, Locale locale) {
        return getMessage(resolvable.getKey(), resolvable.getArguments(), locale, resolvable.getDefaultMessage());
    }

    /**
     * Return the error message for the given BeanValidationException.
     * 
     * @param ex The BeanValidationException to resolve the error message for,
     *      using the error code as resource bundle key
     * @param locale The locale to search the resource bundles for
     * @return The resolved error message
     * @see #getMessage(MessageResolvable, Locale)
     */
    public String getMessage(BeanValidationException ex, Locale locale) {
        StringBuilder message = new StringBuilder();
        message.append(getMessage((MessageResolvable) ex, locale));
        for (BeanValidationException valIssue : ex.getValidationIssues()) {
            message.append(System.getProperty("line.separator"));
            message.append(getMessage((MessageResolvable) valIssue, locale));
        }
        return message.toString();
    }

    /**
     * Set the resource bundle names to search for message keys,
     * the order will be preserved when looking for keys. Take care to yield names of existing bundles,
     * otherwise a MissingResourceException may be thrown during message resolution.
     * Names should be in the form like e.g. 'org.openxma.dsl.platform.i18n.ErrorMessage'
     */
    public void setBundleNames(String[] bundleNames) {
        for (String bundleName : bundleNames) {
            if (bundleName == null || bundleName.trim().length() == 0) {
                throw new IllegalArgumentException();
            }
            bundleName = bundleName.trim();
            if (!this.bundleNames.contains(bundleName)) {
                this.bundleNames.add(bundleName);
            }
        }
    }

    public void setPreferredBundleName(String preferredBundleName) {
        if (this.bundleNames.indexOf(preferredBundleName) > 0) {
            this.bundleNames.remove(preferredBundleName);
        }
        if (!this.bundleNames.contains(preferredBundleName)) {
            this.bundleNames.add(preferredBundleName);
        }
    }

    /**
     * Prepare parameters for output, at least arrays of objects should come meaningful,
     * arrays of primitives are not regarded yet.
     * Overwrite this method if you want to apply user defined formatting.
     * @param params Object array of message parameters
     * @return prepared
     */
    protected Object[] prepareParamsForOutput(Object[] params) {
        Object[] preparedParams = new Object[params.length];
        for (int i = 0; i < params.length; i++) {
            if (params[i] instanceof Object[]) {
                preparedParams[i] = formatParam((Object[]) params[i]);
            } else {
                preparedParams[i] = formatParam(params[i]);
            }
        }
        return preparedParams;
    }

    protected Object formatParam(Object[] param) {
        return Arrays.toString(param);
    }

    protected Object formatParam(Object param) {
        return param;
    }

    protected String getDefaultBundleName() {
        return this.defaultBundleName;
    }

    protected ResourceBundle getResourceBundle(String bundleName, Locale locale) throws MissingResourceException {
        return ResourceBundle.getBundle(bundleName, locale);
    }

    protected String resolveMessage(ResourceBundle bundle, String key, Object[] params) {
        try {
            return MessageFormat.format(bundle.getString(key), params);
        } catch (MissingResourceException ex) {
            return null;
        }
    }

    protected String resolveMessageForBundle(String bundleName, String key, Object[] params, Locale locale) {
        ResourceBundle bundle = getResourceBundle(bundleName, locale);
        if (bundle != null) {
            return resolveMessage(bundle, key, params);
        } else {
            return null;
        }

    }

}
