package org.openxma.dsl.platform.validation;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.springframework.beans.BeanWrapper;
import org.springframework.beans.BeanWrapperImpl;
import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.util.Assert;
import org.springframework.util.ClassUtils;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.DefaultMessageCodesResolver;
import org.springframework.validation.Errors;
import org.springframework.validation.MessageCodesResolver;
import org.springframework.validation.Validator;

import at.spardat.enterprise.exc.SysException;
import at.spardat.enterprise.fmt.FmtFactory;
import at.spardat.enterprise.fmt.FmtParseException;
import at.spardat.enterprise.fmt.IFmt;

public abstract class Validators implements Validator {
    private static final Map<Class<?>, Object> PRIMITIVE_DEFAULT_VALUES = new HashMap<Class<?>, Object>();
    static {
        PRIMITIVE_DEFAULT_VALUES.put(Integer.class, new Integer(0));
        PRIMITIVE_DEFAULT_VALUES.put(Boolean.class, Boolean.FALSE);
        PRIMITIVE_DEFAULT_VALUES.put(Short.class, new Short((short) 0));
        PRIMITIVE_DEFAULT_VALUES.put(Byte.class, new Byte((byte) 0));
        PRIMITIVE_DEFAULT_VALUES.put(Character.class, new Character((char) 0));
        PRIMITIVE_DEFAULT_VALUES.put(Long.class, new Long(0));
        PRIMITIVE_DEFAULT_VALUES.put(Double.class, new Double(0));
        PRIMITIVE_DEFAULT_VALUES.put(Float.class, new Float(0));
    };

    /**
     * 
     * @param object
     *            the object containing the given property
     * @param property
     *            name of the property value to return
     * @return the value of the given property or the primitive default value in
     *         case of null valued wrapper type
     */
    public <T> T getValue(Object object, String property, Class<T> propertyType) {
        Assert.notNull(object, "parameter 'object' must not be null");
        Assert.notNull(property, "parameter 'property' must not be null");
        BeanWrapper beanWrapper = new BeanWrapperImpl(object);
        Object value = beanWrapper.getPropertyValue(property);
        if (value == null && ClassUtils.isPrimitiveWrapper(propertyType)) {
            value = PRIMITIVE_DEFAULT_VALUES.get(propertyType);
        }
        return propertyType.cast(value);
    }

    /**
     * Reject the given string value with the given error code and error
     * arguments if the length is greater than the given maximum.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the field value to check
     * @param maxLength
     *            the maximum (inclusive) field length
     */
    public void rejectIfMaxLength(Errors errors, String name, String value, int maxLength) {
        if (value != null && value.trim().length() > maxLength) {
            errors.rejectValue(name, "field.max.length", new Object[] { name, value, maxLength }, "Attribute '" + name
                    + "' must have at most [" + maxLength + "] characters.");
        }
    }

    /**
     * Reject the given string value with the given error code and error
     * arguments if the length is lower than the given minimum length.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the field value to check
     * @param minLength
     *            the minimum (inclusive) field length
     */
    public void rejectIfMinLength(Errors errors, String name, String value, int minLength) {
        if (value != null && value.trim().length() < minLength) {
            errors.rejectValue(name, "field.min.length", new Object[] { name, value, minLength }, "Attribute '" + name
                    + "' must have at least [" + minLength + "] characters.");
        }
    }

    /**
     * Reject the given field value with the given error code and error
     * arguments if the value is greater than the given max value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param maxValue
     *            the maximum (inclusive) field length
     */
    public void rejectIfMaxValue(Errors errors, String name, Number value, int maxValue) {
        if (value != null && value.intValue() > maxValue) {
            errors.rejectValue(name, "field.max.value", new Object[] { name, value, maxValue }, "Attribute '" + name
                    + "' must not be bigger than [" + maxValue + "].");
        }
    }

    /**
     * Reject the given number value with the given error code and error
     * arguments if the value is lower than the given min value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param minValue
     *            the maximum (inclusive) field length
     */
    public void rejectIfMinValue(Errors errors, String name, Number value, int minValue) {
        if (value != null && value.intValue() < minValue) {
            errors.rejectValue(name, "field.min.value", new Object[] { name, value, minValue }, "Attribute '" + name
                    + "' must not be smaller than [" + minValue + "].");
        }
    }

    /**
     * Reject the given number value with the given error instance if the number
     * of integral digits is greater than the given max value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param maxIntegerDigits
     *            the maximum (inclusive) integral digits for the number
     */
    public void rejectIfMaxIntegerDigits(Errors errors, String name, Number value, int maxIntegerDigits) {
        BigDecimal bigDecimal = toBigDecimal(value);
        int integerPartLength = bigDecimal.precision() - bigDecimal.scale();
        if (integerPartLength > maxIntegerDigits) {
            errors.rejectValue(name, "field.max.integer.digits", new Object[] { name, value, maxIntegerDigits },
                    "Attribute '" + name + "' must have at most [" + maxIntegerDigits
                            + "] digits in the integer portion.");
        }
    }

    /**
     * Reject the given number value with the given error instance if the number
     * of fractional digits is greater than the given max value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param maxFractionalDigits
     *            the maximum (inclusive) fractional digits for the number
     */
    public void rejectIfMaxFractionDigits(Errors errors, String name, Number value, int maxFractionalDigits) {
        BigDecimal bigDecimal = toBigDecimal(value);
        int fractionPartLength = bigDecimal.scale() < 0 ? 0 : bigDecimal.scale();
        if (fractionPartLength > maxFractionalDigits) {
            errors.rejectValue(name, "field.max.fractional.digits", new Object[] { name, value, maxFractionalDigits },
                    "Attribute '" + name + "' must have at most [" + maxFractionalDigits
                            + "] digits in the fraction portion.");
        }
    }

    /**
     * Reject the given number value with the given error code and error
     * arguments if the value is lower than the given min value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the value to check
     * @param format
     *            the format to apply
     */
    public void rejectIfInvalidFormat(Errors errors, String name, String value, String format) {
        rejectIfInvalidFormat(errors, name, value, FmtFactory.create(format, getLocale()));
    }

    /**
     * Reject the given number value with the given error code and error
     * arguments if the value is lower than the given min value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the value to check
     * @param format
     *            the format to apply
     */
    public void rejectIfInvalidFormat(Errors errors, String name, String value, IFmt format) {
        try {
            format.parse(value);
        } catch (FmtParseException fmtParseException) {
            errors.rejectValue(name, getErrorCode(fmtParseException), fmtParseException.getUIMessage(getLocale()));
        }
    }

    /**
     * @param target
     *            target the target bean to validate
     * @return a new errors container
     */
    protected Errors createErrors(Object target) {
        Assert.notNull(target, "parameter 'target' must not be null");
        BeanPropertyBindingResult beanPropertyBindingResult = new BeanPropertyBindingResult(target, target.getClass()
                .getSimpleName());
        beanPropertyBindingResult.setMessageCodesResolver(getMessageCodeResolver());
        return beanPropertyBindingResult;
    }

    /**
     * @return the default <code>MessageCodesResolver</code>
     */
    protected MessageCodesResolver getMessageCodeResolver() {
        return new DefaultMessageCodesResolver();
    }

    /**
     * @return the default <code>Locale</code>
     */
    protected Locale getLocale() {
        return LocaleContextHolder.getLocale();
    }

    /**
     * @param value
     *            the number to convet
     * @return a new BigDecimal instance for the given number value
     */
    protected BigDecimal toBigDecimal(Number value) {
        BigDecimal result = null;
        if (value instanceof BigDecimal) {
            result = (BigDecimal) value;
        } else if (value != null) {
            result = new BigDecimal(value.toString());
        }
        return null != result ? result.stripTrailingZeros() : BigDecimal.ZERO;
    }

    // TODO replace this
    private String getErrorCode(FmtParseException fmtParseException) {
        try {
            return (String) fmtParseException.getClass().getField("bundleKey_").get(fmtParseException);
        } catch (Exception e) {
            throw new SysException(e);
        }
    }
}
